//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// You can use this code under one of two licenses:
//
// This source code is an unpublished proprietary document, and is available 
// for commercial (money making) purposes only under a written license.  If you
// happen to get a copy of this document without agreement, then all you have
// to do is to adhere to the GPL below.
//
// Or,
//
// You can adhere to the GPL, and that will also be okay (in this case,
// no written license is otherwise needed.)  I am not in love with the
// GPL -- and don't mind if the ideas are reused, but prefer that
// my code not be directly used commercially without my knowedge.
// I want to know what is going on with the code.
//
// Code modifications are NOT accepted (unless there is a mutually signed
// written agreement.)  Otherwise, reports, and requests to fix bugs
// are happily received.  (this is for licensing reasons, not for ego reasons.)
//
// A written agreement can modify any part of these copyright/licensing terms
// above.  Such an agreement must be signed/counter-signed by both parties.
// 

//
// This file contains a suprisingly effective and simple audio expansion processor
//
//
// SOME OF THE CAPABILITY OF THIS CODE IS REDUNDANT -- DUE TO AN EARLIER BUG, I HAD THOUGHT
// THAT IT WAS NECESSARY TO RUN MULTI-AXIS decoding, which might still be helpful for audible
// quality -- but not necessary to decode the targeted compression.  Doing additional
// expansion would require optimized algorithms for that purpose, and such code does exist
// in my finalizing processor program (this code was partially derived from the finalizer.)
//
// THERE ARE LOTS OF DIFFERENCES BETWEEN THIS AN MY FINALIZING PROCESSOR.  The finalizer uses
// 8 bands of audio, and not just 4 like this expander.  It also uses the very advanced
// attack/decay code, and has multiple compression/expansion passes tuned to various purposes,
// with a distortion free hard limiter scheme.  Up to 5-6 of various audio processing passes,
// with no gain control intermod (or at least, very-very little!!!)  Some of the attack/release
// sections can be enabled to have FULL waveform following -- which does not sound ideal, but
// with a little additional modification of the gain control -- works IDEALLY.
//

//
// There is excess precision in the calculations here: vectors of double precision
// instead of using floats...  However, the compiler that I use sometimes chokes on 
// non-full-native vectors -- it sometimes produces worse code with vectors of 4 floats instead
// of vectors of 4 doubles.   Since I am doing very few divisions, the cost of doing
// doubles instead of floats on an Intel architecture isn't very significant, but the cost
// of garbled (but correct) code due to using short vectors of floats can be significant.
// None of the large storage buffers use double precision, so memory bandwidth is also
// not excessively impacted (maybe register spills, short term storage and minimal argument
// passing.)
//
// The finalizing processor just uses floats -- and I am considering a reconfiguration of the
// existing 8 bands -- possibly up to 16bands.  Rather than being approx octave width, they would
// be approx 1/2 octave -- that opens up some other interesting possibilities, but that
// is a rather aggressive move -- and might be more trouble than it is worth.  It is already
// tricky to properly handle the 8 bands without overly 'flattening' the spectrum.  I am
// using some good methods that allow the processor to maintain the same spectrum shape without
// creating 'policies' which might change the sound.
//
// HINTS ON READING THE VECTOR CODE:
// * the vector type vtype4 is a vector of 4 double precision floats
//
// * when doing a constructor or type cast that looks like 'val = vtype4(-1.0)', it
//       fills all vector members with the specified value.
//
// * when doing a constructor or type cast that looks like 'val = vtype4({a,b,c,d})',
//       it fills in each element as val[0] = a, val[1] = b, val[2] = c,
//       val[3] = d.
//
// * very important: the vector type descriptor is a type in its own right.  That
//       class type describes all aspects of the integer or float vector type
//       of the same shape.  For example, an vector of 4 32bit floating point
//       and a vector of 4 32bit integers use the same descriptor type.  The
//       descriptor describes the type sizes, the underlying types, etc.
//       The actual vector type is usually either varrayT<VDESC> or IvarrayT<VDESC>.
//       There is also a raw vector type like UvarrayT<VDESC> that remembers the
//       shape, but data that has that type can only do conversions or raw assignments.
//       "vtype4" is a typedef of varrayT<TYPEvd4x1>, which means 'a vector of
//       four doubles with a count of one such HW vector.
//
// * math between vectors of the same type are supported.  All of +,-,*,/,
//       and, or, xor, Vmax, Vmin, Vgt, Vlt, Vge, Vle, (Vselect -- does
//			an ?: type operation on each element of the vector),
//       pow(vec,scalar), pow(vec,vec), log, log10, exp, exp10, dB2lin, lin2dB,
//       ([] -- array access of each element), var.fprint(FILEP, fmt),
//       and probably other things.
//
// * some math between vectors and scalars is very partially supported -- I haven't
//      filled in the combinations that support all of the vector/scalar combos.
//      Multiplication that does 'vector *= scalar' does work, but not much more.
//
// * some math between vector types (e.g. vector of 4 doubles vs. vector of 4 floats) is
//      partially supported.  Assignments ARE supported between types, but are meant to be
//      warned/errored if misused -- so mistakes are attempted to be caught.
//
// * some math between vector types that make little sense almost works (e.g. addition
//      between a vector size of 4 and a vector size of 8.) Expect 'interesting'
//      results until this is debugged.
//
// * no attempt is made to support matrices.
//
// * there are integer analogs of all floating point vector types.  Math between such types
//      is limited due to lack of good semantic definition.
//
// * assignment between integer and floating point vector types does the full type conversion,
//      but no special cases are handled, and there is no unsigned integer type yet.
//
// * there is an idiom for RAW copying between integer and floating point types "intvec = intvectype(fpvec.raw())"
//      which devolves the fp vec type to a raw type (with only conversion operations), and then raw copies
//      to the integer.  There are limited purposes for this, but some of the HW vector
//      opcodes (vblendvps/vblendvpd) encourage a certain mix between integer and FP because of the
//      CPU/HW use of an fp value as a logical bitmask.
//
// * the vector code is designed to support multiple HW vectors as one logical one -- but
//      the multiple vectors aren't fully supported yet.  All of the structure does exist.
//


//
// All timings are for 44.1k -- there are some underlying support mechanisms that handle the conversions
// to other sample rates.
//
// For REAL fun, disassemble this turkey -- the vector code can make it almost impossible to trace without
// experience in working with the vector CPU operations. :-).
//

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>

#include <ctype.h>

#include <malloc.h>
#include <signal.h>

//
// itqA.h contains manifest constants
//
#include "itqA.h"

//
// general typedefs
//
#include "itq-typedefs.h"

//
// inter-thread communicatsion queues/queue management (thread stuff is pretty much dead for
// the expansion processor -- it is too simple to require threading.)
//
//#include "itq-qdefs.h"

//
// miscellaneous filters (IIR/FIR filter code, filter coefficients/etc.)
//
#include "agcfilters.h"
#include "math.h"

//
// Stupid/poorly written low level audio I/O routines
//
#include "audioio.h"

//
// Data recording/gain logging/etc
//
#include "accuminfo.h"

//
// This is a filter class from the very powerful 'smfilt.h' file which contains all kinds of attack/release schemes.
// I have included it here so that the dolbyA code can be kept seperate so that it can be distributed without giving
// away my cool secrets. :-).
//

//
// atkdcys -- create a list of attack/decays supporting the dynamic(mirror) attack/decay
// mechanisms.
//

template <typename DT = double, int NSTEPS = 16>
class atkdcys
{
	DT atks[NSTEPS];
	DT aratios[NSTEPS];
	DT dcys[NSTEPS];
	DT atkst, atkend, atkratio;
	DT dcyst, dcyend, dcyratio;
	float tscale;
	
public:
	atkdcys(float Atscale, DT Aatkst, DT Aatkend, DT Adcyst, DT Adcyend) :
		atks(), dcys(),
		atkst(Aatkst), atkend(Aatkend), atkratio(Aatkend/Aatkst),
		dcyst(Adcyst), dcyend(Adcyend), dcyratio(Adcyend/Adcyst),
		tscale(Atscale)
	{

		tscale = (Atscale < 1.0f) ? 1.0f : Atscale;

		atkst = Aatkst * tscale;
		atkst = Vmin(atkst, DT(1.0));

		atkend = Aatkend * tscale;
		atkend = Vmin(atkend, DT(1.0));

		dcyst = Adcyst * tscale;
		dcyst = Vmin(dcyst, DT(1.0));
	
		dcyend = Adcyend * tscale;
		dcyend = Vmin(dcyend, DT(1.0));

		atkratio = pow(DT(atkend / atkst), (DT(1.0) / DT(NSTEPS-1)));
		dcyratio = pow(DT(dcyend / dcyst), (DT(1.0) / DT(NSTEPS-1)));

		DT atktmp{atkst};
		DT dcytmp{dcyst};

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = atktmp;
			atktmp *= atkratio;
			//aratios[i] = vsqrt(atkend / atktmp);
			aratios[i] = (atkend / atktmp);
			dcys[i] = dcytmp;
			dcytmp *= dcyratio;
		}
	}

	atkdcys(DT Aatkst, DT Aatkend, DT Adcyst, DT Adcyend) :
		atkdcys(1.0f, Aatkst, Aatkend, Adcyst, Adcyend) {}

	atkdcys(const atkdcys& init) :
		atks(), aratios(), dcys(),
		atkst(init.atkst), atkend(init.atkend), atkratio(init.atkratio),
		dcyst(init.dcyst), dcyend(init.dcyend), dcyratio(init.dcyratio),
		tscale(init.tscale)
	{

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = init.atks[i];
			dcys[i] = init.dcys[i];
			aratios[i] = init.aratios[i];
		}
	}

	atkdcys& 
	operator=(const atkdcys& src)
	{
		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = src.atks[i];
			dcys[i] = src.dcys[i];
			aratios[i] = src.aratios[i];
		}
		atkst = src.atkst;
		atkend = src.atkend;
		atkratio = src.atkratio;
		dcyst = src.dcyst;
		dcyend = src.dcyend;
		dcyratio = src.dcyratio;
		tscale = src.tscale;
		
		return *this;
	}

	void
	setatk(int idx, DT val)
	{
		atks[idx] = val;
		aratios[idx] = atkend / val;
	}

	void
	setdcy(int idx, DT val)
	{
		dcys[idx] = val;
	}

	const DT&
	atk( int idx) const
	{
		return atks[idx];
	}

	const DT&
	aratio( int idx) const
	{
		return aratios[idx];
	}

	const DT&
	dcy( int idx) const
	{
		return dcys[idx];
	}

	void 
	retuneatks(DT fromatk, DT toatk)
	{
		float newatkratio = pow(double(toatk / fromatk), double(1.0/(NSTEPS - 1)));
		float atktmp{fromatk};
		atkst = fromatk;
		atkend = toatk;

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = atktmp;
			aratios[i] = toatk / atktmp;
			atktmp *= newatkratio;
		}
	}

	void 
	retunedcys(DT fromdcy, DT todcy)
	{
		float newdcyratio = pow(double(todcy / fromdcy), double(1.0/(NSTEPS - 1)));
		float dcytmp{fromdcy};

		for (int i = 0; i < NSTEPS; i++)
		{
			dcys[i] = dcytmp;
			dcytmp *= newdcyratio;
		}
	}

};

//
// VfilterS is an attack/decay mirror meant to provide both attack/decay intermod
// minimization.  This is similar to Vfilter, but has a slightly different attack
// time specification, and minimizes both attack & decay intermod.
// The attack time works to soften the attack -- minimizing the intermodulation.
// There are other versions of filters in the smfilt.h that don't have the
// soft attack, and are more useful further into the processing.  This
// filter is best used for the first filter after the raw level detection.
//

template <typename VDEF, int NSTEPS=16, int NF=1, bool NEG=false>
class VfilterS
{
	using VT=varrayT<VDEF>;
	using VIT=IvarrayT<VDEF>;
	using FT=typename VDEF::FT;

	atkdcys<VT,NSTEPS> ad;

	VT curval[NF][NSTEPS];
	VT totalv[NF];
	VT initval;
	VT atks, atke;

	VfilterS& operator=(const VfilterS&) = delete;

public:
	typedef VT bsa;
	typedef VIT bsai;
	static constexpr int VS = VDEF::VS;

	VfilterS(VT atkst,VT atkend,
		VT dcyst, VT dcyend, VT Ainitval=FT(0.0f)) :
		ad(atkst, atkend, dcyst, dcyend),
		curval(), totalv(), initval(Ainitval),
		atks(atkst), atke(atkend)
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
			}
			totalv[i] = VT(initval);
		}
	}

	VfilterS(FT atkst, FT atkend,
		FT dcyst, FT dcyend, FT Ainitval=FT(0.0f)) :
		ad(VT(atkst), VT(atkend), VT(dcyst), VT(dcyend)),
		curval(), totalv(), initval(Ainitval),
		atks(VT(atkst)), atke(VT(atkend))
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
			}
			totalv[i] = VT(initval);
		}
	}

	VfilterS(const VfilterS<VDEF,NSTEPS,NF>& init) :
		ad(init.ad), curval(), totalv(), initval(init.initval),
		atks(init.atks), atke(init.atke)
		
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = init.curval[i][j];
			}
			totalv[i] = init.totalv[i];
		}
	}

	~VfilterS()
	{
	}

	VT 
	next(const VT& newval, bool atkdecayall = false)
	{
		for (int j = 0; j < NF; j++)
		{
			VT rtvalX(0.0f);
			static constexpr float LARGEVAL = 1.0e+38;
			static constexpr float SMALLVAL = -1.0e+38;

			if (NEG)
				rtvalX = VT(LARGEVAL);
			else
				rtvalX = VT(SMALLVAL);

			bool atkdecay = (j == 0) || atkdecayall;
			for (int i = 0; i < NSTEPS; i++)
			{
				VT atk = ad.atk(i);
				VT dcy = ad.dcy(i);
				VT Tcv = curval[j][i];

				const VT& Tnewval { (j != 0) ? curval[j-1][i] : newval };

				VT atkflags, dcyflags;

				if (NEG)
				{
					atkflags = Vlt(Tnewval, Tcv);
					dcyflags = Vgt(Tnewval, Tcv);
				}
				else
				{
					atkflags = Vgt(Tnewval, Tcv);
					dcyflags = Vlt(Tnewval, Tcv);
				}

				VT atkamt = atkdecay ? (Tcv * (VT(1.0f) - atk)) + (Tnewval * atk) : Tnewval;
				Tcv = Vselect(atkflags, atkamt, Tcv);
				VT dcyamt = (Tcv * (VT(1.0f) - dcy)) + (Tnewval * dcy);
				Tcv = Vselect(dcyflags, dcyamt, Tcv);

				if (NEG)
					rtvalX = Vmin(rtvalX, Tcv * ad.aratio(i));
				else
					rtvalX = Vmax(rtvalX, Tcv * ad.aratio(i));

				curval[j][i] = Tcv;
			}
			totalv[j] = rtvalX;
		}

		VT rtval = totalv[NF-1];
		if (NEG)
		{
			for (int i = 0; i < NF; i++)
			{
				rtval = Vmin(rtval, totalv[i]);
			}
		}
		else
		{
			for (int i = 0; i < NF; i++)
			{
				rtval = Vmax(rtval, totalv[i]);
			}
		}
			
		return rtval;
	}

	void 
	setcurvals(const VT& initv)
	{
		for (int j = 0; j < NF; j++)
		{
			VT *cv = curval[j];
			for (int i = 0; i < NSTEPS; i++)
			{
				cv[i] = initv;
			}
			totalv[j] = initv;
		}
	}

	void
	setcurvals(FT newval)
	{
		VT nv(0.0f);
		for (int i = 0; i < VDEF::VS; i++)
			nv[i] = newval[i];
		setcurvals(nv);
		return;
	}

	VT
	getcurvals(void) const
	{
		return totalv;
	}

	void
	setatk(int idx, VT newval)
	{
		ad.setatk(idx, newval);
	}

	template <typename PV> 
	PV operator()(const PV& input)
	{
		return next(input);
	}

	template <typename PV> 
	PV operator()(const PV& input, bool atkall)
	{
		return next(input, atkall);
	}

	template <typename PV, typename IV> 
	PV operator()(const PV& input, const IV& aa, const IV& adX)
	{
		return next(input, aa, adX);
	}

	template <typename PV, typename IV> 
	PV operator()(const PV& input, const IV& aa, const IV& adX, bool atkall)
	{
		return next(input, aa, adX, atkall);
	}
};

#undef NOINLINE

//
// If we are running a recent GCC, IF we want 'noinline', we probably
// want 'noclone' also.
//
#if defined(__GNUC__) && (__GNUC__ > 5)
#define NOINLINE __attribute__((noinline,noclone))
#else
#define NOINLINE __attribute__((noinline))
#endif


//
// default initial linear input rms value
//
static constexpr float defaultrms = 1.0f;

//
// default initial dB input value (relative to 1 as 0dB)
//
static constexpr float defaultdB = 0.0f;

//
// The Vector/float types used to be selectable due to problems with the control
// signal filter dynamic ranges.  That problem has been worked around, so we
// use a fixed set of types.  Now, we use the varray series of types, and
// can easily select the types as needed.  Also, dynamic range issues are in the past.
//

//
// Maximum number of espansion steps -- only one is required for kind-of DolbyA decoding,
// since bugfixes make the compression/expansion
// work correctly!!!
//
	static constexpr int MAXEXP = 16;

static Asample
ninetymatrix(const Asample& in, bool mode = false)
{
					if (mode == false)
					{
	               float mid = (in.item0() + in.item1());
  		             float side = (in.item0() - in.item1());

						mid *= 0.281f;
						side *= 0.959f;

						return Asample(mid + side, mid - side) * 0.707f * 1.02f;
					}
					else
					{
						Asample rtv(2.302f*in.item0() + 1.256f*in.item1(), 1.256f * in.item0() + 2.302f*in.item1());
						return rtv * 1.25 * 0.707f;
					}
};

static Asample
msmatrix(const Asample& in, bool mode = false)
{
	
	if (mode == false)
	{
		float mid = (in.item0() + in.item1());
		float side = (in.item0() - in.item1());

		return Asample(mid * 0.707f, side * 0.707f);
	}
	else
	{
		float left = (in.item0() + in.item1());
		float right = (in.item0() - in.item1());
		return Asample (left * 0.707f, right * 0.707f);
	}
}


//
// Manifest constants for configuration expressions handled by the procexpcmd procedure/expcfgT class
//

	static constexpr int DMODE_SKIP = 0;
	static constexpr int DMODE_LR = 1;
	static constexpr int DMODE_90 = 2;
	static constexpr int DMODE_MS = 3;
	static constexpr int DMODE_LRGAINONLY = 4;
	static constexpr int DMODE_90GAINONLY = 5;
	static constexpr int DMODE_MSGAINONLY = 6;
	static constexpr int DMODE_LRWONLY = 7;
	static constexpr int DMODE_90WONLY = 8;
	static constexpr int DMODE_MSWONLY = 9;
	static constexpr int DMODE_LRHFONLY = 10;
	static constexpr int DMODE_90HFONLY = 11;
	static constexpr int DMODE_MSHFONLY = 12;
	static constexpr int DMODE_90RWONLY = 13;

	static constexpr const char * const modeid[15] {
		"SKIP", "LEFT-RIGHT", "IN-BETWEEN", "MID-SIDE",
		"LR-GAINONLY", "BTW-GAINONLY", "MS-GAINONLY",
		"LR-WIDTHONLY", "BTW-WIDTHONLY", "MS-WIDTHONLY",
		"LR-HFONLY", "BTW-HFONLY", "MS-HFONLY",
		"BTW-RWIDTHONLY",
		"NONE"
	};

	bool hfonly(int md)
	{
			return ((md >= DMODE_LRHFONLY) && (md <= DMODE_MSHFONLY));
	}


//
// class/data structure that controls the decoding (expanding).  One of these per step
//
	struct expcfgT
	{
		int dmode;
		float dBthresh;
		float LINthresh;
		float dBgainbefore;
		float LINgainbefore;

		expcfgT() : dmode(DMODE_SKIP), dBthresh(0.0f), LINthresh(1.0f), dBgainbefore(0.0f), LINgainbefore(1.0f) {}

		void setgaindB(float dBgain)
		{
			dBgainbefore = dBgain;
			LINgainbefore = dB2lin(dBgain);
		}

		void setgainLIN(float LINgain)
		{
			LINgainbefore = LINgain;
			dBgainbefore = lin2dB(LINgain);
		}

		void setthreshdB(float dBval)
		{
			dBthresh = dBval;
			LINthresh = dB2lin(dBval);
		}

		void setthreshLIN(float LINval)
		{
			LINthresh = LINval;
			dBthresh = lin2dB(LINval);
		}

		void dump(void) const
		{
			fprintf(stderr,"%10.10s dBthresh: %6.2f, dBgain: %6.2f\n", modeid[dmode], dBthresh, dBgainbefore);
		}
	};

//
// Number of expansion steps
//
	int numexp = 0;

//
// Array containing the steps
//
	expcfgT expcfg[MAXEXP];


//
// dump all of the dolby steps
//
	void 
	dumpallexp(void)
	{
		for (int i = 0; i < numexp; i++)
		{
			fprintf(stderr,"step%d: ", i);
			expcfg[i].dump();
		}
	}

//
// Return whether or not this is a dolby processor or if it is a fixed gain
//
	bool
	isdexpand(int index)
	{
		int dm = expcfg[index].dmode;
		if (dm == DMODE_LR || dm == DMODE_90 || dm == DMODE_MS || dm == DMODE_LRHFONLY || dm == DMODE_90HFONLY || dm == DMODE_MSHFONLY)
			return true;
		return false;
	}

	bool
	isxexpand(int index)
	{
		int dm = expcfg[index].dmode;
		return false;
	}

//
// parses the expansion control commands -- a string array of
// (<cmd>,<thresh>,<gain>), producing the expcfg array, with a count
// of numexp.
//
	int 
	procexpcmd(const char *incmd)
	{
		enum expcmdst {NOST = 0, INCMD = 1, INTHRESH = 3, INGAIN = 5, FINISH=999 };

		int incp = 0;
		char inc;
		expcmdst dcst{NOST};

		static constexpr int MAXTMPBUFSIZE = 128;
		char tmpbuf[MAXTMPBUFSIZE];

		int tmpbufcnt = 0;
		expcfgT* curexpcfg = &expcfg[0];
		numexp = 0;
		
		//
		// trivial parsing states
		//

nextitem:
		while ((dcst != FINISH) && (inc = incmd[incp]) != '\0')
		{
			incp++;
			inc = tolower(inc);

			switch (dcst)
			{
	case NOST:
				if (isblank(inc))
					break;

				if (inc == '(')
				{
					curexpcfg = &expcfg[numexp];

					if (numexp == 0)
						*curexpcfg = expcfgT{};
					else
						*curexpcfg = expcfg[numexp-1];

					numexp++;

					dcst = INCMD;
					tmpbufcnt = 0;
				}
				break;

	case INCMD:
				if (isblank(inc))
					break;

				if (inc == ',' || inc == ')')
				{
					tmpbuf[tmpbufcnt] = '\0';
					if (tmpbufcnt != 0)
					{
						bool gainmode = false;
						bool widthmode = false;
						bool rwidthmode = false;
						bool hfmode = false;
						int tptr = 0;

						char tcmd = tmpbuf[tptr];

						if (tcmd == 'h')
						{
								hfmode = true;
								tptr++;
								tcmd = tmpbuf[tptr];
								if (tcmd == '\0')
									tcmd = 'l';
						}

						if (tcmd == 'g')
						{
								gainmode = true;
								tptr++;
								tcmd = tmpbuf[tptr];
								if (tcmd == '\0')
									tcmd = 'l';
						}

						if (tcmd == 'w')
						{
								widthmode = true;
								tptr++;
								tcmd = tmpbuf[tptr];
								if (tcmd == '\0')
									tcmd = 'l';
						}

						if (tcmd == 'y')
						{
								rwidthmode = true;
								tptr++;
								tcmd = tmpbuf[tptr];
								if (tcmd == '\0')
									tcmd = 'b';
						}

							
						if ((tcmd == 'l') || (tcmd == 'r'))
						{
							curexpcfg->dmode = hfmode ? DMODE_LRHFONLY : (widthmode ? DMODE_LRWONLY : (gainmode ? DMODE_LRGAINONLY : DMODE_LR));
						}
						else if ((tcmd == 'm') || (tcmd == 's'))
						{
							curexpcfg->dmode = hfmode ? DMODE_MSHFONLY : (widthmode ? DMODE_MSWONLY : (gainmode ? DMODE_MSGAINONLY : DMODE_MS));
						}
						else if ((tcmd == 'b') || (tcmd == 'i') || (tcmd == '9'))
						{
							curexpcfg->dmode = rwidthmode ? DMODE_90RWONLY : (hfmode ? DMODE_90HFONLY : (widthmode ? DMODE_90WONLY : (gainmode ? DMODE_90GAINONLY : DMODE_90)));
						}
					}

					tmpbufcnt = 0;

					if (inc == ')')
						dcst = NOST;
					else
						dcst = INTHRESH;
			
					break;
				}

				tmpbuf[tmpbufcnt] = inc;
				tmpbufcnt++;
				break;

	case INTHRESH:
				if (isblank(inc))
					break;

				if (inc == ',' || inc == ')')
				{
					tmpbuf[tmpbufcnt] = '\0';
					if (tmpbufcnt != 0)
					{
						curexpcfg->setthreshdB(atof(&tmpbuf[0]));
					}

					tmpbufcnt = 0;

					if (inc == ')')
						dcst = NOST;
					else
						dcst = INGAIN;

					break;
				}

				tmpbuf[tmpbufcnt] = inc;
				tmpbufcnt++;
				break;

	case INGAIN:
				if (isblank(inc))
					break;

				if (inc == ',' || inc == ')')
				{
					tmpbuf[tmpbufcnt] = '\0';
					if (tmpbufcnt != 0)
					{
						curexpcfg->setgaindB(atof(&tmpbuf[0]));
					}
					tmpbufcnt = 0;

					dcst = NOST;

					if ( inc == ',')
					{
						fprintf(stderr,"comma not expected after the gain parameter at character location %d\n", incp);
					}
					break;
				}
				tmpbuf[tmpbufcnt] = inc;
				tmpbufcnt++;
				break;

	case FINISH:
	default:
				break;
			}
			
		}
		return 0;
	}
		
//
// process and dump the commands
//
void
pdcmd(const char *incmd)
{
	procexpcmd(incmd);
	if (ac.info > 2) dumpallexp();
}


//
// Define the data types.   The vector mechanism is supported by the file varray.h -- and provides
// the various operations/conversion between all kinds of floatingpoint and integer SIMD types
//
#if defined(EIGHTWIDE)
using VDEF4 = TYPEvd4x1;			// this is the descriptor for the vector type
using VFP4 = VDEF4::FT;				// this is the underlying floating point type in the vector
using vtype4 = varrayT<VDEF4>;	// the actual vector type that we are using

using VDEF8 = TYPEvf8x1;			// this is the descriptor for the vector type
using VFP8 = VDEF8::FT;				// this is the underlying floating point type in the vector
using vtype8 = varrayT<VDEF8>;	// the actual vector type that we are using

#else

using VDEF4 = TYPEvf4x1;			// this is the descriptor for the vector type
using VFP4 = VDEF4::FT;				// this is the underlying floating point type in the vector
using vtype4 = varrayT<VDEF4>;	// the actual vector type that we are using

using VDEF8 = TYPEvf8x1;			// this is the descriptor for the vector type
using VFP8 = VDEF8::FT;				// this is the underlying floating point type in the vector
using vtype8 = varrayT<VDEF8>;	// the actual vector type that we are using

#endif

namespace da
{
	//
	// basic attack/decay times derived from lots of listening of apparently appropriately
	// 'compressed' material.
	// the approximate times are 1/(value * 44100)
	//
	static constexpr VFP4 satk = 0.0102;			// LF/MF decay time for fast attacks
	static constexpr VFP4 sdcy = 0.00038;		// LF/MF attack/decay with nonlinear components below
	static constexpr VFP4 fatk = 0.0202;			// HF decay time for fast attacks
	static constexpr VFP4 fdcy = 0.00077;		// HF attack/decay with nonlinear components below


	static constexpr VFP4 hpfsdcy0 = sdcy * 0.34;
	static constexpr VFP4 hpffdcy0 = fdcy * 0.34;

	static constexpr VFP4 hpfrcomp = 0.64;
	//	static constexpr VFP4 hpfrcomp = 0.67;
	static constexpr VFP4 hpfsdcy = sdcy * hpfrcomp;
	static constexpr VFP4 hpffdcy = fdcy * hpfrcomp;

	using VfiltS = VfilterS<VDEF4,48>;  // 48 steps is overdesign, 16 is probably enough.

//
// Expander routine
//
// IN THIS FILE -- ALL FILTER COEFFICIENTS ARE FOR 44.1K, even if the processor doesnt' support 44.1K.  Other
// portions of the code do the conversion for the sample rates.
//
class dbexpand
{

	static constexpr float dBminus5=0.562341;
	static constexpr float dBminus10=0.316228;
	//
	//  Direct or pseduo feedback mode
	//
#define FBMODE 0
static constexpr int MAXGAINHPF=64;

	//
	// general types used for gains/levels for the 4 channels
	//
	// VDEF4 is the varrayT descriptor, while vtype4 (a single vector of 4 VFP4s)
	// is the actual type of the data
	//

	using GAINHPFTYPE = hpfT<vtype4, VFP4>;

	//
	// modified single attack/release control type -- with the nonlinear elements
	// added for either speeding up an overly soft attack, or slowing down an overly
	// quick attack.  This is obviously the scheme used by Dolby to help with
	// avoiding the strong attack glitches (a form of intermod.)
	//
	// I am echoing the type of attack/decay used originally only for the effect on
	// the sound, not to avoid glitches.  I have a proprietary scheme that avoids
	// the 'glitches' or 'crackles' WITHOUT modifying the attack times (my proprietary
	// scheme can work with any attack/decay times without glitching.)
	//

	//
	// gain control/level detect filters -- described below
	//
	VfiltS dfiltlLNB;
	VfiltS dfiltrLNB;


	//
	// threshold offset (lowest level threshold) LINear and dB
	//

	static constexpr VFP4 defaultdBToffset = -41.75;

	VFP4 LINToffset, dBToffset;
	vtype4 EXpvtI;
	vtype4 dBEXpvtI;
	vtype4 INPthreshold; 
	vtype4 dBINPthreshold; 
	vtype4 XINPthreshold; 
	vtype4 XdBINPthreshold; 

	//
	// sample count
	//
	int scnt;

	//
	// high pass filters/DC&LF removal
	//
	hpf4As inputfilt0;
	hpf4As inputfilt1;
	hpf4As outputfilt0;
	hpf4As outputfilt1;
	hpf4As outputfilt2;
	hpf4As outputfilt3;

	//
	// band splitter
	//
	dATsingle infilt;

	//
	// gain passed back
	//
	vtype4 pgainl, pgainr;
	int curgainpf;

	//
	// output cleanup filters (different curves based on sample rates 
	// mostly 'gone' by 30kHz at higher sample rates.)
	//
	lpf20500 ofilt;

//
// Explicit filtering to avoid aliasing -- multiply control signal with audio...
//
	lpf20500 sfilt[4];
	lpf100 gfilt0;
	lpf1200 gfilt1;
	lpf3000 gfilt2;
	lpf3000 gfilt3;

	//
	// index for gain/index for thresholds
	//
	int myindex, mytindex;
	//
	// Special modes
	//
	unsigned mode;

	sVdly4As<512> LRsigdly;

	//
	// nonlinear constants for the visible attack/decay times
	// painfully/iteratively derived.  Surprisingly, the third order
	// below is critical -- otherwise moderate attacks will seem
	// blunted.  Generally, the most critical number is the 2nd order,
	// and seems close to correct -- perhaps needs a little optimization.
	// At this point, the difference between o3 being 0.0 or the value specifed below
	// is more important than a small 5-10 difference in o2.
	//
	// BOTH atko2mult and atko3mult could use a bit more tuning.  They are definitely in the
	// 'ballpark.'
	//

	static constexpr VFP4 atko2mult = 0.00;
	static constexpr VFP4 atko3mult = 0.00;
	static constexpr VFP4 atko4mult = 0.00;
	static constexpr VFP4 atko5mult = 0.00;
	//
	// fast averaging -- really meant to help avoid the noise from
	// any inaccurate or other inconsistency in transcendental functions
	//
//
// these simple inline functions with 'gf' in their names are the gainfunctions.  These
// are the dB/dB conversions that implement the expansion ratios.  This doesn't use linear
// dB/dB expansion, so these formulas do the calcuations.
//
// NOTE THAT THESE FORMULAS ARE EXPERIMENTS AND REQUIRED NUMEROUS LISTENING ITERERATIONS TO DERIVE
// THESE CURVES.   Painfully derived and historical code tests -- there is alot of chaff in these snippets,
// but I don't have the heart to remove them.
//
	template <typename DT>
	DT fullgf(DT lvl)
	{
		return (DT(-0.0) - (DT(0.046986) * lvl)) - (DT(0.006598) * lvl * lvl);
	}

	template <typename DT>
	DT fgfe(DT llvl)
	{
		DT gfval = fullgf(llvl);
		DT extgf = Vselect(Vgt(llvl, DT(-7.0)), DT(-0.50000) + (llvl - DT(-7.0)) * DT(0.100), gfval);
		extgf = Vselect(Vlt(llvl, DT(-40.0)), (DT(-9.17736) + (llvl + DT(40.0))* DT(1.0)), extgf);
		extgf = Vmin(extgf, DT(0.0));
		return extgf;
	}

	VFP4 fgfe(VFP4 llvl)
	{
		vtype4 rtval(fgfe(vtype4(llvl)));
		return rtval[0];
	}

public:
	dbexpand(int index = 2) /* NOINLINE */ :

//
// dfiltlLG/dfiltrLG impliment the quick attack/ msec release component of the expanders general attack/release scheme
//
// in these attack/release schemes, the following describe the actions:
//
// the first line is the release time (ignore the manifest symbols satk/fatk -- they are release
//                times when in the first line)
// the second line is the attack time -- 1.0 is infinitely fast
// the third line is the 2nd order nonlinear component of the attack time
// the 4th line is the 3rd order nonlinear component of the attack time
// if the 3rd or 4th lines are missing, then they are zero -- zero is effectively disabled.
//
//
// These are simple 1st order nonlinear (seperate attack/decay filters) with the following general concept
// for fixed attack/decay times.  The nonlinear versions are a little more complicated (but all encapsulated into
// a single class.)  The method is SOMETHING like this:
//
// if (attack)
//   outputval += (newval - oldoutputval) * attackcoeff;
// else if (decay)
//   outputval += (newval - oldoutputval) * decaycoeff;
//
// There are also some infinitely (almost truly) more complex filter classes available in my audio
// gain control filtering package for very good audio processing.  I am amazed that R Dolby is
// using such simple filters and getting such good results (but his nonlinear
// attack trick is very good to help avoid some of the worst intermod issues!!!!)
//
//
// dfiltlLNB/dfiltrLNB implement the important dynamic attack/fixed release scheme
// This filter works as a dynamic attack time between 1 and 's or fatk' and the
// decay is the sdcy or fdcy fixed.
//
		dfiltlLNB(
			vtype4(1.0),
			vtype4({sratetmod(satk),sratetmod(satk),sratetmod(fatk),sratetmod(fatk)}),
			vtype4({sratetmod(sdcy), sratetmod(sdcy), sratetmod(fdcy), sratetmod(fdcy)}),
			vtype4({sratetmod(sdcy), sratetmod(sdcy), sratetmod(fdcy), sratetmod(fdcy)}),
			vtype4(0.1)),
		dfiltrLNB(
			vtype4(1.0),
			vtype4({sratetmod(satk),sratetmod(satk),sratetmod(fatk),sratetmod(fatk)}),
			vtype4({sratetmod(sdcy), sratetmod(sdcy), sratetmod(fdcy), sratetmod(fdcy)}),
			vtype4({sratetmod(sdcy), sratetmod(sdcy), sratetmod(fdcy), sratetmod(fdcy)}),
			vtype4(0.1)),

		LINToffset(),
		dBToffset(),

		EXpvtI(),
		dBEXpvtI(),
		dBINPthreshold(),
		INPthreshold(),
		XdBINPthreshold(),
		XINPthreshold(),


//
// this is the LF filter removing DC (one for each freq band)  The values are the LF filter coefficients
//
		//inputfilt0(sratetmod(0.00010), sratetmod(0.0002267), sratetmod(0.0004535), sratetmod(0.0004535)),
		//inputfilt1(sratetmod(0.00010), sratetmod(0.0002267), sratetmod(0.0004535), sratetmod(0.0004535)),
		inputfilt0(sratetmod(0.00010), sratetmod(0.0002267), sratetmod(0.01), sratetmod(0.02)),
		inputfilt1(sratetmod(0.00010), sratetmod(0.0002267), sratetmod(0.01), sratetmod(0.02)),
		outputfilt0(sratetmod(0.00010), sratetmod(0.0004), sratetmod(0.02000), sratetmod(0.04000)),
		outputfilt1(sratetmod(0.00010), sratetmod(0.0004), sratetmod(0.02000), sratetmod(0.04000)),
		outputfilt2(sratetmod(0.00010), sratetmod(0.0004), sratetmod(0.02000), sratetmod(0.04000)),
		outputfilt3(sratetmod(0.00010), sratetmod(0.0004), sratetmod(0.02000), sratetmod(0.04000)),
//
// this is the input band splitting filter
//
		infilt(),

//
// sample count
//
		scnt(0),

//
// gain passed back from previous sample
//
		pgainl(1.0), pgainr(1.0),
		curgainpf(0),

		ofilt(),
		sfilt(),
		gfilt0(), gfilt1(), gfilt2(), gfilt3(),
		myindex(index), mytindex(index),
		mode(0U)

	{

			{
				char *cval = getenv("DBAMODE");
				if (cval != nullptr)
				{
					mode = unsigned(strtol(cval, nullptr, 16));
//					if (ac.info) fprintf(stderr,"DBAMODE: 0x%0x\n", mode);
				}
			}

			dBToffset = defaultdBToffset;
			LINToffset = dB2lin(dBToffset);

			EXpvtI = vtype4(expcfg[mytindex].LINthresh);
			dBEXpvtI = vtype4(expcfg[mytindex].dBthresh);

			dBINPthreshold = dBEXpvtI + vtype4(dBToffset);
			INPthreshold = dB2lin(dBINPthreshold);

	}

	struct vtype4pair
	{
		vtype4 v0, v1;
	};

	vtype4pair
	getToffset(float dBTin)
	{
			EXpvtI = vtype4(expcfg[mytindex].LINthresh);
			dBEXpvtI = vtype4(expcfg[mytindex].dBthresh);

			LINToffset = dB2lin(dBTin);
			dBToffset = dBTin;
			vtype4 dBthresh = dBEXpvtI + vtype4({dBToffset, dBToffset, dBToffset - VFP4(0.0), dBToffset - VFP4(8.75)});

			return vtype4pair {
				dBthresh, dB2lin(dBthresh)
			};
	}

	void 
	setToffset(float dBTin)
	{
			vtype4pair v = getToffset(dBTin);

			dBINPthreshold = v.v0;
			INPthreshold = v.v1;

			v = getToffset(dBTin - 10.0);
			//v = getToffset(dBTin);
			XdBINPthreshold = v.v0;
			XINPthreshold = v.v1;
	}

	void
	setindex(int idx)
	{
		myindex = idx;
	}

	void
	settindex(int idx)
	{
		mytindex = idx;
		setToffset(dBToffset);
	}

	static constexpr VFP4 mingain = 0.100;  // this is the number that programs the maximum loss (expansion)

//
// An enterprising person might be able to put this into a simple plugin.   There is alot of code
// that isn't presented here (but is in supporting routines) like audio filters, the actual code
// for attack/decay filters, and the various tricky parts like the gain management code that
// avoids most of the AGC distortions.  I didn't enable the anti-ripple code because I think
// that the compliment of this expander expects that the ripple is present.  However, the
// other kinds of distortion continue to be suppressed (or avoided.)
//
	vtype4
	nextlevel(int dmode, const Asample& siginA, Asample& sigout)
	{
			Asample sigin[4];

			//
			// Take the simple stereo input signal, and split into 4 freq bands as represented
			// by 'sigin'.
			//
			infilt(siginA, sigin);
			inputfilt0(sigin, sigin);
			inputfilt1(sigin, sigin);

			//
			// This is the output signal that is built from the 4 output channels.
			//
			Asample sigouttmp{};

			//
			// This is the 4 channel representation of the output signal
			//
			Asample sigoutbuf[4];

			//
			// ofilt does some removal of any resulting out of band stuff.
			// oddly -- the iir filters are actually much more effective near the
			// highest audio frequencies at the lower sample rates.  However,
			// I suspect that the decoder otherwise generally works best at 96k.
			// 192k requires twice the CPU, and I doubt that there is not
			// much benefit.
			// (at 44.1k samples/sec, the filter falls to -60dB by 21000,
			// while it sits fat and happy at about -35dB at 30kHz with
			// a 96k sample rate -- odd, for otherwise the same filter
			// parameters, heh?  Thats the world of digital filters.)
			//
			// I could have done a really good brickwall at 20500Hz, and just
			// might do so in the future -- but it does require lots of taps,
			// and might slow down the program a bit when running at higher
			// sample rates.  As is, the two filters are only 4th order IIR,
			// elliptical with a very little bit of ripple.  Since there are
			// two filters, I guess that the -35dB at 30kHz becomes -70dB,
			// and the middling frequencies of 25kHz are pretty well suppressed.
			// That is why I did two filters instead of one.  And also, back
			// when I thought that we needed to do the multi-axis version of
			// the expander, I didn't want subsequent steps to be interfered
			// with by the earlier steps.
			//
			// Also, and most importantly -- earlier versions of the code used
			// more real feedback of the actual audio -- and could fall into
			// oscillation if I made a bug, or even if the math on a processor
			// or library might work a little differently.  So, with so very
			// much attenuation at frequencies higher than might be noticed, it
			// might save a tweeter or some other physical device if the program
			// goes nuts.  It is very, very stable now.  The only thing that
			// could POSSIBLY go unstable is the gain control, not the audio
			// itself.
			//

			//
			// Here, we calculate the raw unfiltered signal squared
			// this is mostly just being used for absolute value
			// but in the past, this section of code came from a very
			// nice compressor which used carefully designed RMS
			// measurements.  Most code long gone -- I just do it
			// this way now.
			//
			vtype4 LINrmsinsql(VFP4(1.0));
			vtype4 LINrmsinsqr(VFP4(1.0));

			for (int i = 0; i < 4; i++)
			{
				Asample& stmp = sigin[i];

				float lsig = ((stmp.item0() * stmp.item0()));
				LINrmsinsql[i] = lsig;
				float rsig = ((stmp.item1() * stmp.item1()));
				LINrmsinsqr[i] = rsig;
			}

			//
			// All filters are linear (NOT SQUARED -- unlike most of my tech, where I tend to work
			// in the squared and dB domain.)
			//

			//
			// get the linear values
			//
			vtype4 LINrmsinl = vsqrt(LINrmsinsql);
			vtype4 LINrmsinr = vsqrt(LINrmsinsqr);

			//
			// add in the threshold as determined above
			//
			// we can choose from an additive threshold or a hard
			// threshold
			//
			LINrmsinl += INPthreshold;
			LINrmsinr += INPthreshold;
			//LINrmsinl = Vmax(LINrmsinl, INPthreshold);
			//LINrmsinr = Vmax(LINrmsinr, INPthreshold);

			//
			// Do the attack/release -- the LG filter does a fast attack, and
			// millisecond decay. The LNB filter does the slow attack/slow decay
			// These filters are put together in a contorted fashion.   The reason why
			// these filters are put together this way is based upon the DolbyA behavior.
			// If you change the code between 'DOLBYA FILTER CODE BEGIN/END', the
			// characteristics will not likely match a real DolbyA unit.  I don't normally
			// write my attack/decay filters like this, so there is a real reason for this
			// structure.
			//

// APPROX DOLBYA FILTER TIMING CODE BEGIN

			vtype4 LINrmsrawtmpl = dfiltlLNB(LINrmsinl);
			vtype4 LINrmsrawtmpr = dfiltrLNB(LINrmsinr);


// APPROX DOLBYA FILTER TIMING CODE END


			//
			// get the dB values for the dB calculations -- I derived my expansion curves
			// in the dB domain and haven't done the conversion to linear yet.
			// This is a technically unnecessary waste, but not too bad on recent machines.
			//
			// I have decided to remove the special threshold code (and support for the
			// differing LF/HF compression/expansion curve versions)
			//
			// All we need to do now is to convert the dB/dB gain curve to the
			// linear world.   Curve-fitting time again!!!  I dont' think that it is
			// possible to do the change algebraically (I already tried doing it myself
			// and using Maxima.)  I want to be fairly faithful to the curve -- it took
			// significant amount of time to tweak the curve until it sounded right.  If
			// I let myself, I could probably tweak it until the end of time.
			//
			// removing the lin-to-dB, applying the curve, then dB-to-lin, and just
			// doing a curve should greatly simplify the code and really speed it up.
			//
			vtype4 dBlvll = lin2dB(LINrmsrawtmpl);
			vtype4 dBlvlr = lin2dB(LINrmsrawtmpr);

			//
			// set the level to pivot against
			//
			dBlvll -= dBEXpvtI;
			dBlvlr -= dBEXpvtI;

			//
			// Here, we apply the compression ratio (currently in dB terms)
			//
			vtype4 dBnll = fgfe(dBlvll);
			vtype4 dBnlr = fgfe(dBlvlr);

			
			dBnll = Vmin(dBnll, vtype4(-0.50));
			dBnlr = Vmin(dBnlr, vtype4(-0.50));

			//
			// Convert back to linear
			//
			pgainl = dB2lin(dBnll);
			pgainr = dB2lin(dBnlr);

			//
			// Here we move the level values around for various legacy reasons.  If you are trying to figure out
			// WHY I am moving the values around as much as I am -- don't bother.  I am just doing it because the
			// old code had reasons to do so.  Now, there is a serious need for a code cleanup!!!
			//
			vtype4 dgainl = pgainl;
			vtype4 dgainr = pgainr;

			//
			// Here, we are calculating the gain for 9k-20k
			//
			VFP4 pgainl3 = (pgainl[3] * pgainl[2]) / ((pgainl[3] + pgainl[2]) * VFP4(0.500));
			VFP4 pgainr3 = (pgainr[3] * pgainr[2]) / ((pgainr[3] + pgainr[2]) * VFP4(0.500));

			//
			// Set the 10dB loss limit for the lower three bands
			//
			for (int i = 0; i < 3; i++)
			{
				pgainl[i] = Vmax(pgainl[i], VFP4(dBminus10));
				pgainr[i] = Vmax(pgainr[i], VFP4(dBminus10));
			}

			//
			// set the 15dB loss limit for the highest band
			//
			pgainl[3] = Vmax(pgainl3, VFP4(dBminus10 * dBminus5));
			pgainr[3] = Vmax(pgainr3, VFP4(dBminus10 * dBminus5));


			//
			// gain never higher than 0dB
			//
			pgainl = Vmin(pgainl, vtype4(VFP4(1.0)));
			pgainr = Vmin(pgainr, vtype4(VFP4(1.0)));
					

			//
			// calculate new signal levels for each of the 4 bands
			// 0-80, 80-3k, 3k-9k, 9k-20k
			//
			LRsigdly(sigin);
			const Asample *sigtmpi = LRsigdly(int(aspec.SRATETDIV * 0.75));
			Asample sigtmph[4];
			outputfilt2(sigtmpi, sigtmph);
			outputfilt3(sigtmph, sigtmph);

			sigtmph[2] -= sigtmph[3];

//
// These are the filters for the gain control and signal --
// meant to mitigate unnecessary aliasing.  This isn't such
// a severe problem for analog designs, and this caught me
// off guard.
//
// Note that the left and right control signals are somewhat
// oddly encoded into an 'Asample' data item.  This only
// helps to reuse already written filter code.
//
			{
				Asample g{};
				g.item0() = pgainl[0];
				g.item1() = pgainr[0];
				g = gfilt0(g);
				pgainl[0] = double(g.item0());
				pgainr[0] = double(g.item1());

				g.item0() = pgainl[1];
				g.item1() = pgainr[1];
				g = gfilt1(g);
				pgainl[1] = g.item0();
				pgainr[1] = g.item1();

				g.item0() = pgainl[2];
				g.item1() = pgainr[2];
				g = gfilt2(g);
				pgainl[2] = double(g.item0());
				pgainr[2] = double(g.item1());

				g.item0() = pgainl[3];
				g.item1() = pgainr[3];
				g = gfilt3(g);
				pgainl[3] = double(g.item0());
				pgainr[3] = double(g.item1());
			}

			for (int i = 0; i <= 3; i++)
			{
				//
				// sample 20kHz filter
				//
				sigtmph[i] = sfilt[i](sigtmph[i]);

			}

			if (!hfonly(dmode))
			{
				sigoutbuf[0] = Asample{ sigtmph[0].item0() * float(pgainl[0]), sigtmph[0].item1() * float(pgainr[0]) };
				sigoutbuf[1] = Asample{ sigtmph[1].item0() * float(pgainl[1]), sigtmph[1].item1() * float(pgainr[1]) };
			}
			else
			{
				sigoutbuf[0] = sigtmph[0];
				pgainl[0] = 1.0f;
				pgainr[0] = 1.0f;
				sigoutbuf[1] = sigtmph[1];
				pgainl[1] = 1.0f;
				pgainr[1] = 1.0f;
			}
			sigoutbuf[2] = Asample{ sigtmph[2].item0() * float(pgainl[2]), sigtmph[2].item1() * float(pgainr[2])};
			sigoutbuf[3] = Asample{ sigtmph[3].item0() * float(pgainl[3]), sigtmph[3].item1() * float(pgainr[3])};


			Asample ob[4];

			if (( mode & 0x2) == 0)
			{
				outputfilt0(sigoutbuf, ob);
				outputfilt1(ob, ob);
			}
			else
			{
				ob[0] = sigoutbuf[0];
				ob[1] = sigoutbuf[1];
				ob[2] = sigoutbuf[2];
				ob[3] = sigoutbuf[3];
			}

			//
			// build the output -- summing each of the 4 bands
			//
			if (mode & 0x100)
			{
				for (int i = 1; i < 4; i++)
				{
					sigouttmp += ob[i];
				}
			}
			else if (mode & 0x200)
			{
				for (int i = 0; i < 1; i++)
				{
					sigouttmp += ob[i];
				}
			}
			else
			{
				for (int i = 0; i < 4; i++)
				{
					sigouttmp += ob[i];
				}
			}

			sigout = ofilt(sigouttmp);

			vtype4 avggain = (pgainl + pgainr) * vtype4(VFP4(0.50));
			return avggain;
	}

	void
	next(int dm, const Asample& sigin, Asample& sigout)
	{
			//fprintf(stderr,"<%d,%d>", myindex, dm);
			vtype4 gain = nextlevel(dm, sigin, sigout);

			//
			// looking at the schematic -- the fet has a big resistor in the gate -- enough
			// that it should be accounted for by a little bit of a TC -- I used 1msec.
			//

			DLBYAst[myindex].next(gain);

			scnt++;
			return;
	}
 };

};


//
// This owns and runs the expansion passes -- traversing the control table --
// interpreting it and invoking the passes.
//
class expA
{
	static constexpr int NPASSES = MAXEXP;
	static constexpr int NFRQCHAN = 4;
	static constexpr float gaincomp = 1.000f;
	static constexpr float hfcomp = 0.000f;

	da::dbexpand dbx[NPASSES];
	//
	// This filter is meant to remove above-audio frequencies.  We can encounter >20kHz
	// because this program works up to 192ksamples/sec.
	//
	lpf20500 ifiltlpf;

	//
	// The input high pass filter is meant mostly just for DC removal.
	// There should be a small amount of attenuation at 20Hz -- the
	// 3dB for each of these filters should be somewhere around 5Hz,
	// so the attenuation in any reasonable audio freq should really
	// be nil.  I am using two simple poles just to clean up the
	// unintentional DC offsets and any odd kinds of turntable rumble if that
	// is encountered.  There is a lot of turntable rumble in the
	// 5-20Hz range, but this isn't meant to address that.  The
	// total loss at 20Hz for two filters should be about 0.52dB.
	//
	hpfAs ifilthpf0;
	hpfAs ifilthpf1;

	//
	// The output high pass filter is meant to help remove super low
	// intermod frequencies resulting from any lack of filtering in the
	// actual gain processor.  I am adding a little more attenuation
	// than the input filters -- so there might be a little more loss
	// at 20Hz.  The 3dB here each is about 10Hz.  The loss at 20Hz
	// should be approx 1dB for each filter.  At 30Hz, 0.9dB, and
	// at 50Hz approx 0.4dB.
	//
	hpfAs ofilthpf0;
	hpfAs ofilthpf1;

	Asample
	expstep(int index, const Asample& insamplearg)
	{
		Asample insample(insamplearg);

		insample *= expcfg[index].LINgainbefore;

		Asample outsample{};
		dbx[index].next(expcfg[index].dmode, insample, outsample);

		outsample *= gaincomp;
		return outsample;
	}

public:
	expA() : dbx(), ifiltlpf(),
		ifilthpf0(sratetmod(0.0001125f)), ifilthpf1(sratetmod(0.0001125f)),
		ofilthpf0(sratetmod(0.000225f)), ofilthpf1(sratetmod(0.000225f))

	{
			//
			// set the indices for each of the expansion passes
			//
			for (int i = 0; i < NPASSES; i++)
			{
				dbx[i].setindex(i);
				dbx[i].settindex(i);
			}
	}

	//
	// this routine does all of the passes for each input sample
	//
	int // NOINLINE
	next(int count, const Asample *sigin, Asample *sigout)
	{
		for (int i = 0; i < count; i++)
		{
			Asample insample = ifiltlpf(ifilthpf1(ifilthpf0(sigin[i])));
			Asample insampletmp{};
			int j = 0;
			for (j = 0; j < numexp; j++)
			{
				Asample originsample = insample;
				float mid, side;
				expcfgT& cfg = expcfg[j];
				float gbefore = cfg.LINgainbefore;
//				fprintf(stderr,"(%d)", cfg.dmode);
				switch (cfg.dmode)
				{

		case DMODE_SKIP:
					break;
		case DMODE_LRHFONLY:
		case DMODE_LR:
					insample = expstep(j, insample);
					break;

		case DMODE_LRGAINONLY:
					insample *= gbefore;
					DLBYAst[j].next(gbefore);
					break;

		case DMODE_LRWONLY:
					insample.setitem1(insample.item1() * gbefore);
					DLBYAst[j].next(gbefore);
					break;
					
		case DMODE_MSHFONLY:
		case DMODE_MSGAINONLY:
		case DMODE_MSWONLY:
		case DMODE_MS:
					mid = insample.item0() + insample.item1();
					mid *= 0.707f;
					side = insample.item0() - insample.item1();
					side *= 0.707f;
					insampletmp = Asample{mid, side};

					if ((cfg.dmode == DMODE_MS) || (cfg.dmode == DMODE_MSHFONLY))
					{
						insampletmp = expstep(j, insampletmp);
					}
					else if (cfg.dmode == DMODE_MSWONLY)
					{
						insampletmp.setitem1(insampletmp.item1() * gbefore);
						DLBYAst[j].next(gbefore);
					}
					else
					{
						insampletmp *= gbefore;
						DLBYAst[j].next(gbefore);
					}

					mid = insampletmp.item0() + insampletmp.item1();
					mid *= 0.707f;
					side = insampletmp.item0() - insampletmp.item1();
					side *= 0.707f;

					//insample = (originsample * 0.50f + Asample{mid, side} * 0.50f);
					insample = Asample{mid, side};
					break;
		case DMODE_90HFONLY:
		case DMODE_90GAINONLY:
		case DMODE_90WONLY:
		case DMODE_90RWONLY:
		case DMODE_90:
					insampletmp = ninetymatrix(insample, false);
					if ((cfg.dmode == DMODE_90) || (cfg.dmode == DMODE_90HFONLY))
					{
						insampletmp = expstep(j, insampletmp);
					}
					else if (cfg.dmode == DMODE_90WONLY)
					{
						insampletmp.setitem1(insampletmp.item1() * gbefore);
						DLBYAst[j].next(gbefore);
					}
					else if (cfg.dmode == DMODE_90RWONLY)
					{
						insampletmp.setitem0(insampletmp.item0() * gbefore);
						DLBYAst[j].next(gbefore);
					}
					else
					{
						insampletmp *= gbefore;
						DLBYAst[j].next(gbefore);
					}
					//insample = (originsample * 0.50f + ninetymatrix(insampletmp, true) * 0.50f);
					insample = ninetymatrix(insampletmp, true);
					break;
				}
			}
			sigout[i] = ofilthpf1(ofilthpf0(insample));
		}

		return count;
	}

};

//
// This routine is extreme hackery -- I folded the infrastructure for the audio processor into one thread -- and cobbled one of the
// threads into this monstrosity (use of static variables/etc.)
//
// Should change this into a class -- but no time right now.
//
void
procmusic(paudiochunk pa)
{
		//
		// This 'static' stuff is cr*p, but it works for now.
		//
		static int totalsamples70;
		static int limit99, limit70, limit50, limit25;
		static bool isinit = false;

		static expA* exp = nullptr;

		if (!isinit)
		{
			exp = new expA;
			isinit = true;
		}

		audiopayload* ap = &pa->payload();
		pAsample sp = &ap->samples[0];
		int totalsamples = ap->samplecount;

		for (int i = 0; i < totalsamples; i++)
		{
			const Asample& as = ap->samples[i];
			float sumsq = as.msumsq();
			INst.next( sumsq);
		}

		exp->next(totalsamples, sp, sp);

		for (int i = 0; i < totalsamples; i++)
		{
			const Asample& as = ap->samples[i];
			float sumsq = as.msumsq();
			OUTst.next( sumsq);
			accuminfo::accumulatelevel(as, totalsamples70, limit99, limit70, limit50, limit25);
		}
}

//
// vi: set ts=3:
//

